/*  (c)2000 Franklin Electronic Publishers, Inc. Burlington, NJ.
    All rights reserved. */

#define	CTABLE_CPP	1

#include <assert.h>
#include <malloc.h>
#include <string.h>
#include "gui.h"
#include "guisys.h"


/* ------------------------------------------------------------------------- */
CTable::CTable( U16 id, U16 width, U16 height, TABLEOPTIONS opt = 0 ) :
	CViewable( id, width, height )
{
	v = (struct CvTable *) malloc( sizeof(struct CvTable) );

	numrows = numcols = numvisible = firstvisible = 0;
	v->var_col = 65535;
	col_widths  = (U16 *) malloc( 0 );
	row_heights = 0;
	child_indices = 0;
	v->row_ids = 0;
	v->scrollbar   = new CScrollBar( 0, 0, height );
	v->scrollbar->Hide();
	v->scrollbar->SetPagesize( 1 ); /* For variable-height lines */
	AddChild( v->scrollbar, width-SCROLLBAR_WIDTH, 0 );
	SetForegroundColor( COLOR_GRAY73 );
	v->options = opt;
}


/* ------------------------------------------------------------------------- */
CTable::~CTable()
{
	if (row_heights) {
		free( row_heights );
	}
	if (v->row_ids) {
		free( v->row_ids );
	}
	if (child_indices) {
		free( child_indices );
	}
	free( col_widths );
	free( v );
}


/* ------------------------------------------------------------------------- */
void
CTable::SetFont( const FONT *font )
{
	assert( !"CTable::SetFont should not be used" );
}


/* ------------------------------------------------------------------------- */
void
CTable::SetWidth( U16 width )
{
	U16 j, k;
	CViewable *object;
	S16 dif = width - GetWidth();

	CViewable::SetWidth(width);
	if (dif) {
		/* Width changing - move the scrollbar */
		v->scrollbar->SetX( width - SCROLLBAR_WIDTH );
		if (row_heights) {
			/* Update the column widths */
			assert( v->var_col != 65535 );
			col_widths[v->var_col] += dif;
			for (j = 0; j < numrows; ++j) {
				object = GetCell( j, v->var_col );
				if (object) {
					object->SetWidth( col_widths[v->var_col] );
				}
				for (k = v->var_col+1; k < numcols; ++k) {
					object = GetCell( j, k );
					if (object) {
						object->SetX( object->GetX() + dif );
					}
				}
				UpdateRow( j );
			}
		}
	}
}


/* ------------------------------------------------------------------------- */
void
CTable::SetHeight( U16 height )
{
	U16 dif = height - GetHeight();

	CViewable::SetHeight( height );

	if (dif) {
		/* Update scrollbar */
		v->scrollbar->SetHeight( height );
		Update_Position( firstvisible );
	}
}


/* ------------------------------------------------------------------------- */
void
CTable::GetUsableRect( RECT *rect ) const
{
	CViewable::GetUsableRect( rect );

	rect->x      += 2;
	rect->width  -= 4;
	rect->y      += 2;
	rect->height -= 4;

	if (!v->scrollbar->IsHidden()) {
		assert( rect->width >= SCROLLBAR_WIDTH );
		rect->width -= SCROLLBAR_WIDTH;
	}
}


/* ------------------------------------------------------------------------- */
U16
CTable::GetMaxHeight() const
{
	S32 result = 4; /* Top and bottom border */
	U16 row, sep = 3;
	ROWHEIGHT *heights = row_heights;

	if (v->options & TABLEOPTION_NOROWGRID) {
		sep = 0;
	}
	for (row = 0; row < numrows; ++row) {
		result += heights->max + sep;
		++heights;
	}

	if (result > 65535) {
		return 65535;
	}
	return (U16) result;
}


/* ------------------------------------------------------------------------- */
void
CTable::Draw()
{
	CWindow *window = GetWindow();
	RECT rect;
	U16 j;
	COLOR fgcolor = GetForegroundColor();
	COLOR bgcolor = GetBackgroundColor();

	if (!numvisible) {
		/* First draw for a new table */
		Update_Position(firstvisible);
	}

	if (!window || IsHidden() || GUI_UpdatePending()) {
		/* We'll draw later */
		return;
	}

	GetRect( &rect );
	if (!v->scrollbar->IsHidden()) {
		rect.width -= SCROLLBAR_WIDTH;
	}
	if (!bgcolor) {
		bgcolor = window->GetBackgroundColor();
	}
	window->DrawRectFilled( &rect, bgcolor );
	window->DrawRect( &rect, fgcolor );

	/* Find height of used space and draw row boundaries */
	if (v->options & TABLEOPTION_NOROWGRID) {
		rect.y = GetHeight();
	} else {
		rect.height = 1;
		for (j = 0; j < numvisible; ++j) {
			rect.y += row_heights[firstvisible+j].cur + 3;
			window->DrawRectFilled( &rect, fgcolor );
		}
	}

	if (numcols > 1 && !(v->options & TABLEOPTION_NOCOLGRID)) {
		/* Draw column boundaries */
		rect.height = rect.y - GetY();
		rect.width = 1;
		rect.x = GetX() + 2;
		rect.y = GetY();
		for (j = 0; j < numcols-1; ++j) {
			rect.x += col_widths[j] + 1;
			window->DrawRectFilled( &rect, fgcolor );
			rect.x += 2;
		}
	}
	DrawChildren();
}


/* ------------------------------------------------------------------------- */
S32
CTable::MsgHandler( MSG_TYPE type, CViewable *object, S32 data )
{
	U16 row, col, status;

	//FIXME - handle more cases
	switch (type) {
	case MSG_POSITION:
		if (object != v->scrollbar) {
			/* Some scrollbar in a table cell */
			break;
		}
		status = v->scrollbar->GetStatus();
		if (!(status & SCROLLSTATUS_USER)) {
			if (status & SCROLLSTATUS_INCREMENT) {
				/* Page-down in progress */
				ScrollKey( K_PAGE_DOWN );
			} else if (v->scrollbar->GetStatus() & SCROLLSTATUS_DECREMENT) {
				ScrollKey( K_PAGE_UP );
			} else {
				Update_Position( v->scrollbar->GetPosition() );
			}
		}
		return 1;
	case MSG_NUMLINES_CHANGED:
		row = FindCellRow( object );
		if (row != 65535) {
			UpdateRow( row );
		}
		return 1;
	case MSG_FOCUS:
		if (data == -1 || object == this) {
			return 1;
		}
		if (object == v->scrollbar) {
			//FIXME - scrollbar should handle keystrokes
			return GetWindow()->SetFocus( this );
		}
		row = FindCellRow( object );
		if (row == 65535) {
			/* Some nested object got focus */
			break;
		}
		/* Focus given to table cell that doesn't want it */
		col = FindCellColumn( object );
		if (col == numcols-1) {
			/* Table takes focus */
			return GetWindow()->SetFocus( this );
		}
		return GetWindow()->SetFocus( GetCell(row,col+1) );
	default:
		; /* Fall through */
	}
	return CViewable::MsgHandler( type, object, data );
}

/* ------------------------------------------------------------------------- */
BOOL
CTable::ScrollKey( KEYSYM key )
{
	U16 status, row, rowsep, height;

	row = 0; /* Make compiler happy */

	switch (key) {
	case K_UP:
		v->scrollbar->SetPosition( firstvisible-1 );
		return TRUE;
	case K_DOWN:
		if (firstvisible+numvisible < numrows ||
		    (numrows && row_heights[numrows-1].cur < row_heights[numrows-1].max)) {
			v->scrollbar->SetPosition( firstvisible+1 );
		}
		return TRUE;
	case K_TOP:
		v->scrollbar->SetPosition( 0 );
		return TRUE;
	case K_PAGE_DOWN:
	case K_JOG_DOWN:
		if (firstvisible+numvisible == numrows) {
			/* Nothing to scroll */
			return TRUE;
		}
		row = firstvisible+numvisible-1;
		if (numvisible == 1 ||
		    (row < numrows-1 && row_heights[row].cur == row_heights[row].max)) {
			++row;
		}
		break;
	case K_BOTTOM:
		firstvisible = numrows;
		/* Fall through */
	case K_PAGE_UP:
	case K_JOG_UP:
		/* Page-up in progress */
		if (v->options & TABLEOPTION_NOROWGRID) {
			rowsep = 0;
		} else {
			rowsep = 3;
		}
		row = firstvisible-1;
		if (row == 65535) {
			/* Can't scroll (QA #00682) */
			row = 0;
			break;
		}
		height = GetHeight()-4-rowsep - (row_heights[row].max+rowsep);
		while (row > 0 && height >= row_heights[row-1].max+rowsep) {
			--row;
			height -= row_heights[row].max + rowsep;
		}
		break;
	default:
		return FALSE;
	}
	status = v->scrollbar->GetStatus();
	v->scrollbar->SetStatus( status | SCROLLSTATUS_USER );
	v->scrollbar->SetPosition( row );
	v->scrollbar->SetStatus( status );
	Update_Position( v->scrollbar->GetPosition() );
	return TRUE;
}


/* ------------------------------------------------------------------------- */
void
CTable::AddColumn( U16 width )
{
	assert( !row_heights );
	if (!width) {
		assert( v->var_col == 65535 );
		v->var_col = numcols;
	}
	++numcols;
	col_widths = (U16 *) realloc( col_widths, numcols*sizeof(U16) );
	col_widths[numcols-1] = width;
}


/* ------------------------------------------------------------------------- */
U16
CTable::GetColumnWidth( U16 column ) const
{
	assert( column < numcols );
	return col_widths[column];
}


/* ------------------------------------------------------------------------- */
void
CTable::SetNumRows( U16 nrows )
{
	RECT rect;
	U16 j, width, colsep;

	/* This can be called only once */
	assert( !row_heights );
	numrows = nrows;

	/* Set size of variable-column, or determine that widths all add up */
	GetUsableRect( &rect );
	if (numcols) {
		if (v->options & TABLEOPTION_NOCOLGRID) {
			colsep = 0;
		} else {
			colsep = 3;
		}
		if (numcols == 1) {
			/* No column-dividers for one-column table */
			rect.width += colsep;
		}
		for (j = 0; j < numcols; ++j) {
			width = col_widths[j];
			width += colsep; /* Column divider */
			assert( width <= rect.width );
			rect.width -= width;
		}
	} else {
		/* Table has no columns - give it one */
		col_widths = (U16 *) realloc( col_widths, sizeof(U16) );
		col_widths[0] = 0;
		numcols = 1;
		v->var_col = 0;
	}
	if (v->var_col == 65535) {
		assert( !rect.width );
	} else {
		col_widths[v->var_col] = rect.width;
	}

	/* Finalize the arrays */
	row_heights = (ROWHEIGHT *) calloc( numrows, sizeof(ROWHEIGHT) );
	v->row_ids = (S32 *) calloc( numrows, sizeof(S32) );
	child_indices = (U16 *) calloc( numrows*numcols, sizeof(U16) );

	v->scrollbar->SetMaximum( numrows );
}


/* ------------------------------------------------------------------------- */
void
CTable::InsertRow( U16 nextrow )
{
	assert( row_heights );

	if (nextrow > numrows) {
		nextrow = numrows;
	}
	row_heights = (ROWHEIGHT *) realloc( row_heights,
					     (numrows+1) * sizeof(ROWHEIGHT) );
	memmove( &row_heights[nextrow+1],
		 &row_heights[nextrow],
		 (numrows-nextrow) * sizeof(ROWHEIGHT) );
	memset( &row_heights[nextrow], 0, sizeof(ROWHEIGHT) );

	v->row_ids = (S32 *) realloc( v->row_ids, (numrows+1) * sizeof(S32) );
	memmove( &v->row_ids[nextrow+1],
		 &v->row_ids[nextrow],
		 (numrows-nextrow) * sizeof(S32) );
	v->row_ids[nextrow] = 0;

	child_indices = (U16 *) realloc( child_indices,
					 (numrows+1) * numcols * sizeof(U16) );
	memmove( &child_indices[(nextrow+1)*numcols],
		 &child_indices[nextrow*numcols],
		 (numrows-nextrow) * numcols * sizeof(U16) );
	memset( &child_indices[nextrow*numcols], 0, numcols*sizeof(U16) );

	++numrows;
	v->scrollbar->SetMaximum( numrows );

	if (nextrow < firstvisible) {
		v->scrollbar->SetPosition(firstvisible+1);
	} else if (nextrow <= firstvisible+numvisible) {
		Update_Position( firstvisible );
	}
}


/* ------------------------------------------------------------------------- */
void
CTable::DeleteRow( U16 row, BOOL deletecells = TRUE )
{
	U16 j;

	for (j = 0; j < numcols; ++j) {
		ClearCell( row, j, deletecells );
	}

	--numrows;
	v->scrollbar->SetMaximum( numrows );

	memmove( &row_heights[row],
		 &row_heights[row+1],
		 (numrows-row) * sizeof(ROWHEIGHT) );
	row_heights = (ROWHEIGHT *) realloc( row_heights,
					     numrows*sizeof(ROWHEIGHT) );
	memmove( &v->row_ids[row], &v->row_ids[row+1],
		 (numrows-row) * sizeof(S32) );
	v->row_ids = (S32 *) realloc( v->row_ids, numrows*sizeof(S32) );

	memmove( &child_indices[row*numcols],
		 &child_indices[(row+1)*numcols],
		 (numrows-row) * numcols * sizeof(U16) );
	child_indices = (U16 *) realloc( child_indices, 
					 numrows * numcols * sizeof(U16) );

	if (row < firstvisible) {
		v->scrollbar->SetPosition( firstvisible-1 );
	} else if (row <= firstvisible+numvisible) {
		Update_Position( firstvisible );
		if (row == numrows) {
			/* Deleting anything other than the last row will
			   automatically update the table because the Y
			   coordinates of objects will be changed.  When
			   deleting the last row, we redraw here in case
			   nothing else changed. */
			Draw();
		}
	}
}


/* ------------------------------------------------------------------------- */
S16
CTable::RowY( U16 row ) const
{
	U16 cur;
	S16 result;

	assert( row < numrows );

	if (row < firstvisible) {
		return -32768;
	}
	if (row >= firstvisible+numvisible) {
		return GetY()+GetHeight();
	}
	result = GetY() + 2;
	cur = firstvisible;
	while (cur < row) {
		result += row_heights[cur].cur;
		if (!(v->options & TABLEOPTION_NOROWGRID)) {
			result += 3;
		}
		++cur;
	}

	return result;
}


/* ------------------------------------------------------------------------- */
U16
CTable::RowHeight( U16 row ) const
{
	assert( row < numrows );

	return row_heights[row].cur;
}


/* ------------------------------------------------------------------------- */
U16
CTable::RowMinHeight( U16 row ) const
{
	assert( row < numrows );

	return row_heights[row].min;
}


/* ------------------------------------------------------------------------- */
U16
CTable::RowMaxHeight( U16 row ) const
{
	assert( row < numrows );

	return row_heights[row].max;
}


/* ------------------------------------------------------------------------- */
S32
CTable::GetRowID( U16 row ) const
{
	assert( row < numrows );
	return v->row_ids[row];
}


/* ------------------------------------------------------------------------- */
void
CTable::SetRowID( U16 row, S32 id )
{
	assert( row < numrows );
	v->row_ids[row] = id;
}


/* ------------------------------------------------------------------------- */
void
CTable::UpdateRow( U16 row )
{
	U16 j, base, index, height, rowsep;
	BOOL heightchg = FALSE;
	CViewable *object;
	U16 orig = row_heights[row].max;
	ROWHEIGHT rowheight = { row_heights[row].cur, 0, 0 };

	base = row * numcols;
	for (j = 0; j < numcols; ++j) {
		index = child_indices[base+j];
		if (index) {
			object = GetChild(index);
			height = object->GetMinHeight();
			if (!height) {
				height = object->GetHeight();
			}
			if (rowheight.min < height) {
				rowheight.min = height;
			}
			height = object->GetMaxHeight();
			if (!height) {
				height = object->GetHeight();
			} else if (object->GetHeight() != rowheight.cur) {
				heightchg = TRUE;
			}
			if (rowheight.max < height) {
				rowheight.max = height;
			}
		}
	}

	if (rowheight.cur < rowheight.min) {
		rowheight.cur = rowheight.min;
		heightchg = TRUE;
	}
	if (rowheight.cur > rowheight.max) {
		rowheight.cur = rowheight.max;
		heightchg = TRUE;
	}
	row_heights[row] = rowheight;

	if ((!heightchg && rowheight.max == orig) ||
	    row < firstvisible ||
	    row >= firstvisible+numvisible) {
		/* Changed row is not currently visible or hasn't changed */
		return;
	}

	if (v->options & TABLEOPTION_NOROWGRID) {
		rowsep = 0;
	} else {
		rowsep = 3;
	}

	/* The height of a row currently visible on-screen has been changed */
	base = firstvisible;
	height = GetHeight()-4-rowsep;
	while (row > base && height >= row_heights[row-1].max+rowsep) {
		--row;
		height -= row_heights[row].max + rowsep;
	}
	if (row != firstvisible) {
		v->scrollbar->SetPosition( row );
	} else {
		Update_Position( row );
	}
}


/* ------------------------------------------------------------------------- */
void
CTable::SetCell( CViewable *object, U16 row, U16 column = 0 )
{
	S16 x, y;
	U16 base, sep;

	ClearCell( row, column );

	if (row < firstvisible) {
		object->Hide();
		y = -32768;
	} else if (row >= firstvisible+numvisible) {
		object->Hide();
		y = GetY()+GetHeight();
	} else {
		object->Show();
		y = GetY()+2;
		if (v->options & TABLEOPTION_NOROWGRID) {
			sep = 0;
		} else {
			sep = 3;
		}
		base = firstvisible;
		while (base < row) {
			y += row_heights[base].cur + sep;
			++base;
		}
		if (!(v->options & TABLEOPTION_TOPALIGN)) {
			y += (row_heights[row].cur - object->GetHeight()) >> 1;
		}
	}
	x = GetX()+2;
	if (v->options & TABLEOPTION_NOCOLGRID) {
		sep = 0;
	} else {
		sep = 3;
	}
	for (base = 0; base < column; ++base) {
		x += col_widths[base] + sep;
	}

	child_indices[row*numcols+column] = AddChild( object, x, y );
	if (column == v->var_col) {
		object->SetWidth( col_widths[column] );
	}
	UpdateRow(row);
}


/* ------------------------------------------------------------------------- */
void
CTable::ClearCell( U16 row, U16 column = 0, BOOL deletecell = TRUE )
{
	U16 index, rowcol, *ptr, *end;
	CViewable *object;

	assert( row < numrows && column < numcols );

	rowcol = row*numcols + column;

	index = child_indices[rowcol];
	if (!index) {
		return;
	}
	child_indices[rowcol] = 0;

	object = GetChild(index);
	DeleteChild(index);

	if (deletecell) {
		delete object;
	}

	/* Decrement all indices greater than the one just deleted */
	end = &child_indices[numrows*numcols];
	for (ptr = child_indices; ptr < end; ++ptr) {
		if (*ptr > index) {
			--*ptr;
		}
	}
	UpdateRow( row );
}


/* ------------------------------------------------------------------------- */
CViewable *
CTable::GetCell( U16 row, U16 column = 0 ) const
{
	U16 index = row*numcols + column;

	if (row >= numrows || column >= numcols) {
		return 0;
	}
	index = child_indices[index];
	if (!index) {
		return 0;
	}
	return GetChild(index);
}


/* ------------------------------------------------------------------------- */
U16
CTable::FindRowFromY( S16 y ) const
{
	U16 row;

	assert( row_heights );

	if (y < GetY()+2) {
		/* Y too small */
		return 65535;
	}
	y -= GetY()+2;
	for (row = firstvisible; row < firstvisible+numvisible; ++row) {
		if (y < row_heights[row].cur) {
			return row;
		}
		y -= row_heights[row].cur;
		if (!(v->options & TABLEOPTION_NOROWGRID)) {
			y -= 3;
		}
	}

	/* Y too large */
	return 65535;
}

/* ------------------------------------------------------------------------- */
U16
CTable::FindColFromX( S16 x ) const
{
	U16 col;

	assert( row_heights );

	if (x < GetX()+2) {
		/* X too small */
		return 65535;
	}
	x -= GetX()+2;
	if (!numcols) {
		return -(x >= GetWidth()-4);
	}
	for (col = 0; col < numcols; ++col) {
		if (x < col_widths[col]) {
			return col;
		}
		x -= col_widths[col];
		if (!(v->options & TABLEOPTION_NOCOLGRID)) {
			x -= 3;
		}
	}

	/* X too large */
	return 65535;
}

/* ------------------------------------------------------------------------- */
U16
CTable::FindCellRow( CViewable *object ) const
{
	U16 index = FindChild( object );
	U16 *ptr, *end;

	if (index == 65535 || !index) {
		/* Is not a child or is the scrollbar */
		return 65535;
	}

	end = &child_indices[numrows * numcols];
	for (ptr = child_indices; ptr < end; ++ptr) {
		if (*ptr == index) {
			return (ptr - child_indices) / numcols;
		}
	}

	assert( !"Missing child index" );
	return 65535;
}


/* ------------------------------------------------------------------------- */
U16
CTable::FindCellColumn( CViewable *object ) const
{
	U16 index = FindChild( object );
	U16 *ptr, *end;

	if (index == 65535 || !index) {
		/* Is not a child or is the scrollbar */
		return 65535;
	}

	end = &child_indices[numrows * numcols];
	for (ptr = child_indices; ptr < end; ++ptr) {
		if (*ptr == index) {
			return (ptr - child_indices) % numcols;
		}
	}

	assert( !"Missing child index" );
	return 65535;
}


/* ------------------------------------------------------------------------- */
void
CTable::SetTopRow( U16 row )
{
	assert( row < numrows );

	if (!v->scrollbar->IsHidden()) {
		v->scrollbar->SetPosition( row );
	}
}


/* ------------------------------------------------------------------------- */
void
CTable::HideScrollBar()
{
	U16 j, k, index, *ptr;
	CViewable *child;

	if (v->scrollbar->IsHidden() || !row_heights) {
		return;
	}

	v->scrollbar->Hide();

	col_widths[v->var_col] += SCROLLBAR_WIDTH;
	for (j = 0; j < numrows; ++j) {
		ptr = &child_indices[j*numcols+v->var_col];
		index = *ptr++;
		if (index) {
			GetChild(index)->SetWidth( col_widths[v->var_col] );
		}
		for (k = v->var_col+1; k < numcols; ++k) {
			index = *ptr++;
			if (index) {
				child = GetChild(index);
				child->SetX( child->GetX() + SCROLLBAR_WIDTH );
			}
		}
		UpdateRow( j );
	}
}


/* ------------------------------------------------------------------------- */
void
CTable::ShowScrollBar()
{
	U16 j, k, index, *ptr;
	CViewable *child;

	if (!v->scrollbar->IsHidden()) {
		return;
	}

	v->scrollbar->Show();

	if (!row_heights) {
		return;
	}

	assert( v->var_col != 65535 );
	col_widths[v->var_col] -= SCROLLBAR_WIDTH;
	for (j = 0; j < numrows; ++j) {
		ptr = &child_indices[j*numcols+v->var_col];
		index = *ptr++;
		if (index) {
			GetChild(index)->SetWidth( col_widths[v->var_col] );
		}
		for (k = v->var_col+1; k < numcols; ++k) {
			index = *ptr++;
			if (index) {
				child = GetChild(index);
				child->SetX( child->GetX() - SCROLLBAR_WIDTH );
			}
		}
		UpdateRow( j );
	}
}


/* ------------------------------------------------------------------------- */
BOOL
CTable::IsScrollBarHidden() const
{
	return v->scrollbar->IsHidden();
}


/* ----------------------------------------------------------------------------
   After the scrollbar moves, recalculate the sizes and positions and hidden
   status of the objects.
*/
void
CTable::Update_Position( U16 row )
{
	U16 j, pos, maxheight, rowheight, rowsep, rowofs;
	U16 lastvisible = firstvisible+numvisible-1;
	S32 totheight;
	CViewable *object;
	BOOL need_scrollbar;

	if (v->options & TABLEOPTION_NOROWGRID) {
		rowsep = 0;
	} else {
		rowsep = 3;
	}

	if (firstvisible > row) {
		firstvisible = row;
	} else {
		while (firstvisible < row) {
			/* Rows have scrolled off the top of the screen */
			for (j = 0; j < numcols; ++j) {
				object = GetCell( firstvisible, j );
				if (object) {
					object->Hide();
					object->SetY( -32768 );
				}
			}
			++firstvisible;
		}
	}

	pos = GetY() + 2;
	totheight = GetHeight()-1 - rowsep;
	while (row<numrows && (rowheight = row_heights[row].max) < totheight) {
		/* This row is on screen */
		if (row_heights[row].cur < rowheight) {
			row_heights[row].cur = rowheight;
		}			
		for (j = 0; j < numcols; ++j) {
			object = GetCell( row, j );
			if (!object) {
				continue;
			}
			object->Hide();
			maxheight = object->GetMaxHeight();
			if (maxheight && object->GetHeight() != maxheight) {
				object->SetHeight( maxheight );
			}
			if (v->options & TABLEOPTION_TOPALIGN) {
				rowofs = 0;
			} else {
				rowofs = (row_heights[row].cur - object->GetHeight()) >> 1;
			}
			object->SetY( pos+rowofs );
			object->Show();
		}
		pos += rowheight+rowsep;
		totheight -= rowheight+rowsep;
		++row;
	}

	if (row < numrows && row_heights[row].min < totheight) {
		/* There's room for another row, by squishing it */
		if (row_heights[row].cur != totheight) {
			row_heights[row].cur = totheight;
		}
		for (j = 0; j < numcols; ++j) {
			object = GetCell( row, j );
			if (!object) {
				continue;
			}
			object->Hide();
			maxheight = object->GetMaxHeight();
			if (maxheight && maxheight > totheight) {
				maxheight = totheight;
			}
			if (maxheight && object->GetHeight() != maxheight) {
				object->SetHeight( maxheight );
			}
			if (v->options & TABLEOPTION_TOPALIGN) {
				rowofs = 0;
			} else {
				rowofs = (row_heights[row].cur - object->GetHeight()) >> 1;
			}
			object->SetY( pos+rowofs );
			object->Show();
		}
		++row;
	}

	pos = row;
	while (pos < numrows && pos <= lastvisible) {
		/* These rows have scrolled off the bottom of the screen */
		for (j = 0; j < numcols; ++j) {
			object = GetCell( pos, j );
			if (object) {
				object->Hide();
				object->SetY( GetY() + GetHeight() );
			}
		}
		++pos;
	}
	       
	numvisible = row - firstvisible;

	need_scrollbar = firstvisible ||
	   		 numvisible != numrows ||
                         (numrows > 1 &&
			  row_heights[numrows-1].max !=
			  row_heights[numrows-1].cur);
	if (v->scrollbar->IsHidden()) {
		if (need_scrollbar) {
			ShowScrollBar();
		}
	} else if (!need_scrollbar) {
		HideScrollBar();
	}
} /* Update_Position */


/* Placeholders for future virtual functions */
void CTable::PlaceholderT_K() {};
void CTable::PlaceholderT_L() {};
void CTable::PlaceholderT_M() {};
void CTable::PlaceholderT_N() {};
void CTable::PlaceholderT_O() {};
void CTable::PlaceholderT_P() {};
void CTable::PlaceholderT_Q() {};
void CTable::PlaceholderT_R() {};
void CTable::PlaceholderT_S() {};
void CTable::PlaceholderT_T() {};
void CTable::PlaceholderT_U() {};
void CTable::PlaceholderT_V() {};
void CTable::PlaceholderT_W() {};
void CTable::PlaceholderT_X() {};
void CTable::PlaceholderT_Y() {};
void CTable::PlaceholderT_Z() {};
void CTable::PlaceholderT_AA() {};
void CTable::PlaceholderT_AB() {};
void CTable::PlaceholderT_AC() {};
void CTable::PlaceholderT_AD() {};
void CTable::PlaceholderT_AE() {};
void CTable::PlaceholderT_AF() {};
